KeywordType cKeyWords[] = {
"__asm", "_asm", "asm", "auto", "__automated", "break", "bool", "case", "catch",
"__cdecl", "_cdecl", "cdecl", "char", "class", "__classid", "__closure",
"const", "const_cast", "continue", "__declspec", "default", "delete",
"__dispid", "do", "double", "dynamic_cast", "else", "enum", "__except",
"explicit", "_export", "__export", "extern", "false", "__fastcall", "_fastcall",
"__finally", "float", "for", "friend", "goto", "if", "__import", "_import",
"inline", "", "int", "__int8", "__int16", "__int32", "__int64", "long",
"mutable", "namespace", "new", "operator", "__pascal", "_pascal", "pascal",
"private", "protected", "__property", "public", "__published", "register",
"reinterpret_cast", "return", "__rtti", "short", "signed", "sizeof", "static",
"static_cast", "__stdcall", "_stdcall", "struct", "switch", "template", "this",
"__thread", "throw", "true", "__try", "try", "typedef", "typename", "typeid",
"union", "using", "unsigned", "virtual", "void", "volatile", "wchar_t", "while"
};

#define NCKey      (sizeof(cKeyWords) / sizeof(KeywordType))
#define begin_CKey cKeyWords
#define end_CKey   (& cKeyWords[NCKey])

#if UseSet
KeywordSet CppKeywords(begin_CKey, end_CKey);
#endif

class CppParser : public Parser
{
  public:
    CppParser();
    void DoKeyword(  char C, string& StringC);
    void DoNumber(   char C, string& StringC);
    void DoCharConst(char C, string& StringC);
    void DoComment(  char C, string& StringC);
    void DoPragma(   char C, string& StringC);
    void DoUnknown(  char C, string& StringC);

}; // CppParser

CppParser::CppParser()
#if UseSet
         : Parser(true, & CppKeywords)
#else
         : Parser(true, begin_CKey, end_CKey)
#endif
{
  Handler[Keyword]   = (TypeHandler) & CppParser::DoKeyword;
  Handler[Number]    = (TypeHandler) & CppParser::DoNumber;
  Handler[CharConst] = (TypeHandler) & CppParser::DoCharConst;
  Handler[Comment]   = (TypeHandler) & CppParser::DoComment;
  Handler[Pragma]    = (TypeHandler) & CppParser::DoPragma;
  Handler[Unknown]   = (TypeHandler) & CppParser::DoUnknown;
}

// Keyword - not used

void CppParser::DoKeyword(  char C, string& StringC)
{
  assert(false);
} // CppParser::DoKeyword

// Number - find end of number

void CppParser::DoNumber(   char C, string& StringC)
{
  // This routine assumes that the source is syntactically valid - it does not
  // detect illegal char situations that the BCB3 syntax highlighter does, like
  // 03L or 3.3E. It does not detect illegalities like 3.3E5L, 77L88, or 09.
  // (Otoh, it only accepts X after a single 0, and A..F after an 0X - silly.)
  char UpC = UpCase(C);
  if (isdigit(C) || UpC == 'L' || C == '.' || UpC == 'E' ||
      (UpC == 'X' && TokenText == "0") ||
      (UpC >= 'A' && UpC <= 'F' && TokenText[0] == '0' &&
       UpCase(TokenText[1]) == 'X') ) TokenText += C; // still in number
  else {
    OutputAndReset(Number, TokenText);
    Output(Normal, StringC);
  }
} // CppParser::DoNumber

// CharConst - find end of string/char const

void CppParser::DoCharConst(char C, string& StringC)
{
  TokenText += C;
  if (C == LookFor[0]) {
    int QuoteCount = 0;
    for ( int Index = TokenText.length() - 2;
          Index >= 1 && TokenText[Index] == '\\'; Index-- ) QuoteCount++;
    if ((QuoteCount & 1) == 0) OutputAndReset(CharConst, TokenText);
   }
} // CppParser::DoCharConst

// Comment - verify that '/' starts comment / find end of comment

void CppParser::DoComment(  char C, string& StringC)
{ 
  if (TokenText.length() == 1) {
    TokenText += C;
    switch (C) {
      case '/':
      case '*':
        LookFor   = (C == '/') ? "\r" : "*/";
        LookCount = 0;
        return;
      default:
        OutputAndReset(Normal, TokenText);
    }
  } else {
    bool Append = true;
    bool NewLn = false;
    if (C == LookFor[LookCount++] || (NewLn = C=='\n' && LookFor=="\r") == true) {
      if (LookCount == LookFor.length() || NewLn) {
        bool LineBreak = C == '\r' || C == '\n';
        if (! LineBreak) TokenText += C;
        OutputAndReset(Comment, TokenText);
        Append = LineBreak;
      }
    } else LookCount = 0;
    if (Append) TokenText += C;
  }
} // CppParser::DoComment

// pragma - find end of pragma

bool MultiLinePragma(string& S)
{
  for (int Index = S.length() - 1; Index >= 0; Index--) {
    char C = S[Index];
    if (! iswspace(C)) return C == '\\';
  }
  return false;
} // MultiLinePragma

void CppParser::DoPragma(   char C, string& StringC)
{
  if ((C == '\r' || C == '\n') && ! MultiLinePragma(TokenText))
    OutputAndReset(Pragma, TokenText);
  TokenText += C;
/*
  if ((C == '\r' || C == '\n') && ! MultiLinePragma(TokenText)) {
    OutputAndReset(Pragma, TokenText);
    Output(Normal, StringC);
  } else TokenText += C;
*/
} // CppParser::DoPragma

// Unknown - classify token

void CppParser::DoUnknown(  char C, string& StringC)
{
  if (IsAlpha(C))      { XSet(Normal, StringC); return; }
  else if (isdigit(C)) { XSet(Number, StringC); return; }
  // We can't get here (we've already returned) if C is NOT punctuation
  switch (C) {
    case '/':
      XSet(Comment, StringC);
      return;
    case '#':
      XSet(Pragma, StringC);
      return;
    case '\'':
    case '"':
      XSet(CharConst, LookFor = StringC);
      return;
  }
  TokenText += C; // white space
} // CppParser::DoUnknown

