 //////////////////////////////////////////////////////////////////////////////
//
// Syntax Highlighter, v1.0. Copyright  1998 Midnight Beach, all rights
// reserved. Reads C++ and Pascal source from the clipboard, writes it back
// syntax highlighted in HTML and RTF. Feel free to use the executable however
// you like, without any royalties or licensing. Please do NOT redistribute
// the executable, though - I would like the sole source to stay
// "http://www.midnightbeach.com/ftp_page/SyntaxHighlighter.html". Similarly,
// feel free to modify the source as you wish, but please do not redistribute
// the original source or any changes which you make. Questions and/or comments
// to jon@midnightbeach.com
//

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "CtrlPanel.h"
#include "assert.h"
#include <registry.hpp>
// #include <clipbrd.hpp>
#include "PopupForm.h"
#include "AboutForm.h"

using std::auto_ptr;

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TControlPanel *ControlPanel;

//---------------------------------------------------------------------------

__fastcall TControlPanel::TControlPanel(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------

//
// globals
//

#define TrayID 0x12261958
#define FG "FG"

enum {fmNone, fmBeep, fmPopup} FeedbackMode = fmBeep;
bool         TimedPopup    = true;
unsigned int PopupDuration = 10; // seconds

//
// Create/destroy &c
//

void __fastcall TControlPanel::FormCreate(TObject *Sender)
{
  // Tray icon
  NOTIFYICONDATA Tray;
  Tray.cbSize = sizeof(Tray);
  Tray.hWnd   = Handle;
  Tray.uID    = TrayID;
  Tray.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
  Tray.uCallbackMessage = wm_TrayMsg;
  Tray.hIcon  = Application->Icon->Handle;
  strcpy(Tray.szTip, "Syntax Highlighter - Source to HTML && RTF");
  Shell_NotifyIcon(NIM_ADD, & Tray);

  // font options
  for (int Index = 0; Index < 16; Index++) ColorPanel[Index] = NULL;
  for (int Index = 0, Count = ComponentCount; Index < Count; Index ++) {
    TPanel* Panel = dynamic_cast<TPanel*>(Components[Index]);
    if (Panel != NULL && Panel->Caption == FG) {
      int RGB = ColorToRGB(Panel->Color);
      RGB = ((RGB & 0x00FFFFFF) ^ 0x00FFFFFF);
      Panel->Font->Color = (TColor) RGB;
      Panel->Caption = "";
      ColorPanel[Panel->Tag] = Panel;
    }
  }
  for (int Index = 0; Index < 16; Index++) assert(ColorPanel[Index] != NULL);
  CurrentElement = Low_TokenTypes;

  // Other setup
  Language = Cpp;    // default ....
  LoadPreferences(); // get user preferences
  NoFeedback->Checked        = FeedbackMode == fmNone;
  Beep->Checked              = FeedbackMode == fmBeep;
  ShowFormattedText->Checked = FeedbackMode == fmPopup;
  ShowTime->Position         = (short) PopupDuration;
  ForNSeconds->Checked       = TimedPopup;
  UntilUserClosesIt->Checked = ! TimedPopup;
  SetPopupEnables();

  Application->ShowMainForm = false;
  Shutdown = false;
  RtfOnClipboard = RegisterClipboardFormat("Rich Text Format");

  Setup = true;
  SetElement(0);
} // TControlPanel::FormCreate

void __fastcall TControlPanel::FormDestroy(TObject *Sender)
{
  NOTIFYICONDATA Tray;
  Tray.cbSize = sizeof(Tray);
  Tray.hWnd   = Handle;
  Tray.uID    = TrayID;
  Shell_NotifyIcon(NIM_DELETE, & Tray);

  SavePreferences();
} // TControlPanel::FormDestroy

const AnsiString RegistryKey = "Software\\Midnight Beach\\SyntaxHighlight";
const AnsiString Subkey      = "Settings";

void TControlPanel::LoadPreferences()
{
  // Read from registry
  auto_ptr<TRegistry> Registry(new TRegistry);
  Registry->RootKey = HKEY_CURRENT_USER;

  if (Registry->OpenKey(RegistryKey, false)) {
    auto_ptr<TMemoryStream> Stream(new TMemoryStream);

    int StreamSize = Registry->GetDataSize(Subkey);
    Stream->SetSize(StreamSize);
    Registry->ReadBinaryData(Subkey, Stream->Memory, StreamSize);

    // Read from stream
    Stream->Position = 0;

      Stream->ReadBuffer(SyntaxFormats, sizeof(SyntaxFormats));

      Stream->ReadBuffer(& fLanguage,     sizeof(fLanguage));
      Stream->ReadBuffer(& FeedbackMode,  sizeof(FeedbackMode));
      Stream->ReadBuffer(& TimedPopup,    sizeof(TimedPopup));
      Stream->ReadBuffer(& PopupDuration, sizeof(PopupDuration));

      assert(Stream->Position == Stream->Size);
  } // if OpenKey
  SetLanguage(fLanguage);
} // TControlPanel::LoadPreferences

void TControlPanel::SavePreferences()
{
  // Write to stream
  auto_ptr<TMemoryStream> Stream(new TMemoryStream);

    Stream->WriteBuffer(SyntaxFormats, sizeof(SyntaxFormats));

    Stream->WriteBuffer(& fLanguage,     sizeof(fLanguage));
    Stream->WriteBuffer(& FeedbackMode,  sizeof(FeedbackMode));
    Stream->WriteBuffer(& TimedPopup,    sizeof(TimedPopup));
    Stream->WriteBuffer(& PopupDuration, sizeof(PopupDuration));

  // Write to registry
  auto_ptr<TRegistry> Registry(new TRegistry);
  Registry->RootKey = HKEY_CURRENT_USER;
  if (Registry->OpenKey(RegistryKey, true))
    Registry->WriteBinaryData(Subkey, Stream->Memory, Stream->Size);
} // TControlPanel::SavePreferences

//
// Tray message handler
//

TMenuItem* GetDefaultItem(TMenuItem* Items)
{
  TMenuItem* Result;
  for (int Index = 0, Count = Items->Count; Index < Count; Index++) {
    Result = Items->Items[Index];
    if (Result->Default) return Result;
  }
  return NULL;
} // GetDefaultItem

void __fastcall TControlPanel::wmTrayMsg(TMessage& Message)
{
  assert(Message.WParam == TrayID);
  switch (Message.LParam) {
    case WM_RBUTTONUP:
      SetForegroundWindow(Handle);
      POINT P; GetCursorPos(& P);
      TrayMenu->Popup(P.x, P.y);
      PostMessage(Handle, WM_NULL, 0, 0);
      break;
    case WM_LBUTTONDBLCLK:
      assert(GetDefaultItem(TrayMenu->Items) != NULL);
      SetForegroundWindow(Handle);
      GetDefaultItem(TrayMenu->Items)->OnClick(NULL);
      break;
  }
} // TControlPanel::wmTrayMsg

//
// Misc menu handlers
//

void TControlPanel::SetLanguage(InputLanguages Value)
{
  fLanguage = Value;
  CppMenu->Checked   = Value == Cpp;
  DelphiMnu->Checked = Value == ObjectPascal;
} // TControlPanel::SetLanguage

void __fastcall TControlPanel::CppMenuClick(TObject *Sender)
{
  Language = Cpp;
} // TControlPanel::CppMenuClick

void __fastcall TControlPanel::DelphiMnuClick(TObject *Sender)
{
  Language = ObjectPascal;
} // TControlPanel::DelphiMnuClick

void __fastcall TControlPanel::ExitMnuClick(TObject *Sender)
{
  Shutdown = true;
  Close();
} // TControlPanel::ExitMnuClick

void __fastcall TControlPanel::FormClose(TObject *Sender,
      TCloseAction &Action)
{
  if (! Shutdown) {
    Action = caNone;
    Hide();
  }
} // TControlPanel::FormClose

void __fastcall TControlPanel::AboutMnuClick(TObject *Sender)
{
  auto_ptr<TAbout> About(new TAbout(this));
  About->ShowModal();
} // TControlPanel::AboutMnuClick

void __fastcall TControlPanel::OptionsMnuClick(TObject *Sender)
{
  Show();
} // TControlPanel::OptionsMnuClick

//
// Options UI
//

void __fastcall TControlPanel::FormShow(TObject *Sender)
{
  if (Language == Cpp) CppSource->Checked = true;
  else DelphiSource->Checked = true;
  ShowSampleCode();
  Element->ItemIndex = 0;
  SetElement(0);
} // TControlPanel::FormShow

// element

void __fastcall TControlPanel::ElementClick(TObject *Sender)
{
  SetElement(Element->ItemIndex);
} // TControlPanel::ElementClick

void TControlPanel::SetElement(int N)
{
  if (! Setup) return;

  if (N != CurrentElement)
    ColorPanel[ColorPanelIndex(SyntaxFormats[CurrentElement].Color)]->Caption =
      "";
  CurrentElement = TokenTypes(N);
  SyntaxFormat Format = SyntaxFormats[N];
  ColorPanel[ColorPanelIndex(Format.Color)]->Caption = FG;
  Bold->Checked       = Format.Bold;
  Italic->Checked     = Format.Italic;
  Underline->Checked  = Format.Underline;
} // TControlPanel::SetElement

// color

int TControlPanel::ColorPanelIndex(TColor Color)
{
  Color = (TColor) ColorToRGB(Color);
  for (int Index = 0; Index < 16; Index++)
    if (ColorToRGB(ColorPanel[Index]->Color) == Color) return Index;
  assert(false);
  return -1;
} // TControlPanel::ColorPanelIndex

void __fastcall TControlPanel::ColorPanelClick(TObject *Sender)
{
  assert(dynamic_cast<TPanel*>(Sender)!= NULL);
  TPanel* Panel = (TPanel*) Sender;
  if (SyntaxFormats[CurrentElement].Color != Panel->Color) {
    ColorPanel[ColorPanelIndex(SyntaxFormats[CurrentElement].Color)]->Caption =
      "";
    Panel->Caption = FG;
    SyntaxFormats[CurrentElement].Color = Panel->Color;
    UpdateSampleCode();
  }
}

// Text attributes

void __fastcall TControlPanel::BoldClick(TObject *Sender)
{
  if (SyntaxFormats[CurrentElement].Bold != Bold->Checked) {
    SyntaxFormats[CurrentElement].Bold = Bold->Checked;
    UpdateSampleCode();
  }
} // TControlPanel::BoldClick

void __fastcall TControlPanel::ItalicClick(TObject *Sender)
{
  if (SyntaxFormats[CurrentElement].Italic != Italic->Checked) {
    SyntaxFormats[CurrentElement].Italic = Italic->Checked;
    UpdateSampleCode();
  }
} // TControlPanel::ItalicClick

void __fastcall TControlPanel::UnderlineClick(TObject *Sender)
{
  if (SyntaxFormats[CurrentElement].Underline != Underline->Checked) {
    SyntaxFormats[CurrentElement].Underline = Underline->Checked;
    UpdateSampleCode();
  }
} // TControlPanel::UnderlineClick

// source

void __fastcall TControlPanel::SampleSourceSelect(TObject *Sender)
{
  ShowSampleCode();
} // TControlPanel::SampleSourceSelect

void SetRichText(TRichEdit* RTF, HGLOBAL Global, bool FreeGlobal = true)
{
  std::auto_ptr<TMemoryStream> Stream(new TMemoryStream);

  char* CString = (char*) GlobalLock(Global);
  int Length = strlen(CString); // GlobalSize may be too high
  Stream->SetSize(Length);
  memcpy(Stream->Memory, CString, Length);
  GlobalUnlock(Global);
  if (FreeGlobal) GlobalFree(Global);

  RTF->MaxLength = Length;
  bool PlainText = RTF->PlainText;
  RTF->PlainText = false;
  RTF->Lines->LoadFromStream(Stream.get());
  RTF->PlainText = PlainText;
} // SetRichText

void TControlPanel::ShowSampleCode()
{
  static AnsiString SampleCode[NLanguages] = {
  // AnsiString CppSrc =
  "#define Foo\n"
  "\n"
  "void Foo(bool Numeric)\n"
  "{\n"
  "  if (Numeric) Bar(88 + 0xF00);\n"
  "  else Bar(\"text\");\n"
  "} // Foo\n"
  ,
  // AnsiString DelphiSrc =
  "{$define Foo}\n"
  "\n"
  "procedure Foo(Numeric: boolean);\n"
  "begin\n"
  "  if Numeric\n"
  "    then NumericBar(42 + $F00 * 3.3E5)\n"
  "    else CharBar('text');\n"
  "end; // Foo\n"
  };

  ValidSource = false;

  InputLanguages Language = (CppSource->Checked) ? Cpp : ObjectPascal;
//  AnsiString Src = SampleCode[Language];
  SetRichText( SourceWindow,
               Highlight(RTF, SampleCode[Language].c_str(), Language));
//  SetRichText(SourceWindow, GlobalToString(Highlight(RTF, Src, Language)));

  ValidSource = true;
} // TControlPanel::ShowSampleCode

void __fastcall TControlPanel::SourceWindowSelectionChange(TObject *Sender)
{
  // Normal=N, Keyword=K, Number=#. Char=C, Comment=/, Pragma=P
  static AnsiString SampleMap[NLanguages] = {
  //AnsiString CppMap =
  "PPPPPPPPPPP\r\n"
  "\r\n"
  "KKKKNNNNNKKKKNNNNNNNNN\r\n"
  "N\r\n"
  "NNKKNNNNNNNNNNNNNNN##NNN#####NN\r\n"
  "NNKKKKNNNNNCCCCCCNN\r\n"
  "NN//////\r\n"
  ,
  //AnsiString DelphiMap =
  "PPPPPPPPPPPPP\r\n"
  "\r\n"
  "KKKKKKKKKNNNNNNNNNNNNNNNNNNNNNNN\r\n"
  "KKKKK\r\n"
  "NNKKNNNNNNNN\r\n"
  "NNNNKKKKNNNNNNNNNNNN##NNN####NNN#####N\r\n"
  "NNNNKKKKNNNNNNNNNCCCCCCNN\r\n"
  "KKKNN//////\r\n"
  };

  if (! ValidSource) return;

  InputLanguages Language = (CppSource->Checked) ? Cpp : ObjectPascal;
  char Key = SampleMap[Language][SourceWindow->SelStart];
  if (Key == '\r' || Key == '\n') Key = 'N';
  static string Map = "NK#C/P"; // Normal, Keyword, Numeric, Text, Comment, Pragma
  SetElement(Map.find(Key));
  Element->ItemIndex = (int) CurrentElement;
} // TControlPanel::SourceWindowSelectionChange

void TControlPanel::UpdateSampleCode()
{
  int SelStart = SourceWindow->SelStart;
  ShowSampleCode();
  SourceWindow->SelStart = SelStart;
} // TControlPanel::UpdateSampleCode

// feedback

void __fastcall TControlPanel::FeedbackClick(TObject *Sender)
{
  if (Sender == NoFeedback) FeedbackMode = fmNone;
  else if (Sender == Beep)  FeedbackMode = fmBeep;
  else if (Sender == ShowFormattedText) FeedbackMode = fmPopup;

  SetPopupEnables();
} // TControlPanel::FeedbackClick

void TControlPanel::SetPopupEnables()
{
  bool Popup = FeedbackMode == fmPopup;
  FormattedTextOptionsPnl->Enabled = Popup;
  for ( int Index = 0, Count = FormattedTextOptionsPnl->ControlCount;
        Index < Count; Index++)
    FormattedTextOptionsPnl->Controls[Index]->Enabled = Popup;
} // TControlPanel::SetPopupEnables

void __fastcall TControlPanel::SetShowTimeWidgets(TObject *Sender)
{
  TimedPopup = ForNSeconds->Checked;
  ShowTime->Enabled = TimedPopup;
  ShowTime->Enabled = TimedPopup;
} // TControlPanel::SetShowTimeWidgets

void __fastcall TControlPanel::ShowTimeChanging(TObject *Sender,
      bool &AllowChange)
{
  PopupDuration = ShowTime->Position;
} // TControlPanel::ShowTimeChanging

//
// The main event handler!
//

void TControlPanel::SetIcon(TIcon* Icon)
{
  // Tray icon
  NOTIFYICONDATA Tray;
  Tray.cbSize = sizeof(Tray);
  Tray.hWnd   = Handle;
  Tray.uID    = TrayID;
  Tray.uFlags = NIF_ICON;
  Tray.hIcon  = Icon->Handle;
  Shell_NotifyIcon(NIM_MODIFY, & Tray);
} // TControlPanel::SetIcon

/*
void __fastcall TControlPanel::ConvertMnuClick(TObject *Sender)
{
  Popup->Visible = false;
  SetIcon(BusyIcon->Picture->Icon);

  OpenClipboard(Handle);

  HGLOBAL HTML, RTF, ClipText = GetClipboardData(CF_TEXT);
  char* Source = (char*) GlobalLock(ClipText);
  BothFormats(HTML, RTF, Source, Language);
  GlobalUnlock(ClipText);

  EmptyClipboard(); // establish ownership
  SetClipboardData(CF_TEXT,        HTML);
  SetClipboardData(RtfOnClipboard, RTF);

  CloseClipboard();

//  assert(IsClipboardFormatAvailable(RtfOnClipboard));
//  assert(IsClipboardFormatAvailable(CF_TEXT));

  switch (FeedbackMode) {
    case fmBeep:
      MessageBeep(MB_ICONASTERISK);
      break;
    case fmPopup:
      Popup->ShowHighlitSource( (TimedPopup) ? PopupDuration : INFINITE);
      break;
  }

  SetIcon(Application->Icon);
} // TControlPanel::ConvertMnuClick
*/

void __fastcall TControlPanel::ConvertMnuClick(TObject *Sender)
{
  Popup->Visible = false;
  TWaitCursor Hourglass;
  SetIcon(BusyIcon->Picture->Icon);

  OpenClipboard(Handle);

  HGLOBAL ClipText = GetClipboardData(CF_TEXT);
  Parse((char*) GlobalLock(ClipText), Language);
  GlobalUnlock(ClipText);

  EmptyClipboard(); // establish ownership
  SetClipboardData(CF_TEXT,        NULL); // Delayed rendering
  SetClipboardData(RtfOnClipboard, NULL); // Delayed rendering

//ShowMessage("Haven't CloseClipboard()ed yet");
  CloseClipboard();
//ShowMessage("CloseClipboard");

  switch (FeedbackMode) {
    case fmBeep:
      MessageBeep(MB_ICONASTERISK);
      break;
    case fmPopup:
      Popup->ShowHighlitSource( (TimedPopup) ? PopupDuration : INFINITE);
      break;
  }

  SetIcon(Application->Icon);
} // TControlPanel::ConvertMnuClick

// Clipboard "delayed rendering" message handlers

AnsiString ClipboardFormatName(unsigned int Format)
{
  if (Format == CF_TEXT) return "CF_TEXT";

  char Name[256];
  GetClipboardFormatName(Format, Name, sizeof(Name) - 1);
  return (AnsiString) Name;
}

void __fastcall TControlPanel::wmRenderFormat(TMessage& Message)
{
  unsigned int Format = Message.WParam;
//SetForegroundWindow(Handle);
//ShowMessage("RenderFormat: " + ClipboardFormatName(Format));
  if (Format == CF_TEXT)             SetClipboardData(Format, RenderToHtml() );
  else if (Format == RtfOnClipboard) SetClipboardData(Format, RenderToRTF()  );
//  else ShowMessage("Format request: " + ClipboardFormatName(Format));
  Message.Result = 0;
} // TControlPanel::wmRenderFormat

void __fastcall TControlPanel::wmRenderAllFormats(TMessage& Message)
{
//SetForegroundWindow(Handle);
//ShowMessage("RenderAllFormats");
  SetClipboardData(CF_TEXT,        RenderToHtml() );
  SetClipboardData(RtfOnClipboard, RenderToRTF()  );
  Message.Result = 0;
} // TControlPanel::wmRenderAllFormats

void __fastcall TControlPanel::wmDestroyClipboard(TMessage& Message)
{
//SetForegroundWindow(Handle);
//ShowMessage("DestroyClipboard");
  DiscardParseData();
  Message.Result = 0;
} // TControlPanel::wmDestroyClipboard

