KeywordType pKeyWords[] = {
"and", "exports", "mod", "shr", "array", "file", "nil", "string", "as",
"finalization", "not", "stringresource", "asm", "finally", "object", "then",
"begin", "for", "of", "threadvar", "case", "function", "or", "to", "class",
"goto", "out", "try", "const", "if", "packed", "type", "constructor",
"implementation", "procedure", "unit", "destructor", "in", "program", "until",
"dispinterface", "inherited", "property", "uses", "div", "initialization",
"raise", "var", "do", "inline", "record", "while", "downto", "interface",
"repeat", "with", "else", "is", "resourcestring", "xor", "end", "label", "set",
"except", "library", "shl", "absolute", "external", "pascal", "safecall",
"abstract", "far", "private", "stdcall", "assembler", "forward", "protected",
"stored", "automated", "index", "public", "virtual", "cdecl", "message",
"published", "write", "default", "name", "read", "writeonly", "dispid", "near",
"readonly", "dynamic", "nodefault", "register", "export", "override", "resident"
};

#define NPKey      (sizeof(pKeyWords) / sizeof(KeywordType))
#define begin_PKey pKeyWords
#define end_PKey   (& pKeyWords[NPKey])

#if UseSet
KeywordSet DelphiKeywords(begin_PKey, end_PKey);
#endif

class DelphiParser : public Parser
{
  public:
    DelphiParser();
    void DoKeyword(  char C, string& StringC);
    void DoNumber(   char C, string& StringC);
    void DoCharConst(char C, string& StringC);
    void DoComment(  char C, string& StringC);
    void DoPragma(   char C, string& StringC);
    void DoUnknown(  char C, string& StringC);
}; // DelphiParser

DelphiParser::DelphiParser()
#if UseSet
         : Parser(false, & DelphiKeywords)
#else
         : Parser(false, begin_PKey, end_PKey)
#endif
{
  Handler[Keyword]   = (TypeHandler) & DelphiParser::DoKeyword;
  Handler[Number]    = (TypeHandler) & DelphiParser::DoNumber;
  Handler[CharConst] = (TypeHandler) & DelphiParser::DoCharConst;
  Handler[Comment]   = (TypeHandler) & DelphiParser::DoComment;
  Handler[Pragma]    = (TypeHandler) & DelphiParser::DoPragma;
  Handler[Unknown]   = (TypeHandler) & DelphiParser::DoUnknown;
} // DelphiParser::DelphiParser

// Keyword - not used

void DelphiParser::DoKeyword(  char C, string& StringC)
{
  assert(false);
} // DelphiParser::DoKeyword

// Number - find end of number

void DelphiParser::DoNumber(   char C, string& StringC)
{
  // This routine validates integers - it only allows A..F in hex strings - but
  // it accepts all sort of float point crap - "3..3", "3E" and the like.
  // Formatting code that doesn't compile will have unpredicatble results ....
  char UpC = UpCase(C);
  if (isdigit(C) || C == '.' || UpC == 'E' ||
      (UpC >= 'A' && UpC <= 'F' && TokenText[0] == '$')) TokenText += C;
                                                         // still in number
  else {
    OutputAndReset(Number, TokenText);
    Output(Normal, StringC);
  }
} // DelphiParser::DoNumber

// CharConst - find end of string/char const

void DelphiParser::DoCharConst(char C, string& StringC)
{
  // If C is nonquote, and follows a quote that's not in position 0, output
  // TokenText, reset, and recursively DoUnknown. Append all other chars to
  // TokenText.
  bool Append = C == '\'';
  if (! Append) { // a nonquote
    int Length = TokenText.length();
    Append = Length < 2 || TokenText[Length - 1] != '\'';
  }
  if (Append) TokenText += C;
  else {
    OutputAndReset(CharConst, TokenText);
    DoUnknown(C, StringC);
  }
} // DelphiParser::DoCharConst

// Comment - verify that '/' starts comment / find end of comment

void DelphiParser::DoComment(  char C, string& StringC)
{
  if (LookFor == "") {
    if (TokenText == "{") LookFor = '}';
    else if (TokenText == "/" && C == '/') LookFor = '\r';
         else if (TokenText == "(" && C == '*') LookFor = "*)";
              else { OutputAndReset(Normal, TokenText);
                     DoUnknown(C, StringC);
                     return; }
    TokenText += C;
    if (LookFor != "" && LookFor != "\r" && C == '$') {
    //                   ^^^^^^^^^^^^^^^ no //$ pragmas!
      SetV(Pragma, TokenText);
      return;
    }
  }
  else {
    bool Append = true;
    bool NewLn = false;
    if (C == LookFor[LookCount++] || (NewLn = C=='\n' && LookFor=="\r") == true) {
      if (LookCount == LookFor.length() || NewLn) {
        Append = C == '\r' || C == '\n';
        if (! Append) TokenText += C;
        OutputAndReset(Comment, TokenText);
      }
    } else LookCount = 0;
    if (Append) TokenText += C;
  }
} // DelphiParser::DoComment

// pragma - find end of pragma

void DelphiParser::DoPragma(   char C, string& StringC)
{
  bool Append = true;
  if (C == LookFor[LookCount++]) {
    if (LookCount == LookFor.length()) {
      TokenText += C;
      OutputAndReset(Pragma, TokenText);
      Append = false;
    }
  } else LookCount = 0;
  if (Append) TokenText += C;
} // DelphiParser::DoPragma

// Unknown - classify token

void DelphiParser::DoUnknown(  char C, string& StringC)
{
  if (IsAlpha(C))                  { XSet(Normal, StringC); return; }
  else if (isdigit(C) || C == '$') { XSet(Number, StringC); return; }
  // We can't get here (we've already returned) if C is NOT punctuation or whitespace
  switch (C) {
    case '/':
    case '{':
    case '(':
      LookFor   = "";
      LookCount = 0;
      XSet(Comment, StringC);
      return;
    case '\'':
      XSet(CharConst, StringC);
      return;
  }
  TokenText += C; // white space
} // DelphiParser::DoUnknown


