class TokenToHtml : public TokenToOutput
{
  public:
    TokenToHtml(OutputList& S) : TokenToOutput(S) {};
    virtual void SetFormat(Token& T);
    virtual void ResetFormat(Token& T);
    virtual string DoQuoting(string& S);
}; // TokenToHtml

string HtmlNameOfTColor(TColor Color)
{
  // Get Windows RGB value
  int RGB = ColorToRGB(Color);
  // Convert to HTML BGR value
  RGB = ((RGB & 0x000000FF) << 16) |
        (RGB  & 0x0000FF00) |
        ((RGB & 0x00FF0000) >> 16);
  return IntToString(RGB, 6, 16);
} // HtmlNameOfTColor

// set/reset are in opposite order, so toggles are 'popped from stack' -
// <b><i> ... </i></b>, not <b><i> ... </b></i>

void TokenToHtml::SetFormat(Token& T)
{
  static string Bold      = "<b>";
  static string Italic    = "<i>";
  static string Underline = "<u>";

  SyntaxFormat Format = SyntaxFormats[T.Type];
  if (Format.Bold)      Output(Bold);
  if (Format.Italic)    Output(Italic);
  if (Format.Underline) Output(Underline);
  if (Format.Color != clWindowText)
    Output("<font color=" + HtmlNameOfTColor(Format.Color) + '>');
} // TokenToHtml::SetFormat

void TokenToHtml::ResetFormat(Token& T)
{
  static string UnFont      = "</font>";
  static string UnBold      = "</b>";
  static string UnItalic    = "</i>";
  static string UnUnderline = "</u>";

  SyntaxFormat Format = SyntaxFormats[T.Type];
  if (Format.Color != clWindowText) Output(UnFont);
  if (Format.Underline)             Output(UnUnderline);
  if (Format.Italic)                Output(UnItalic);
  if (Format.Bold)                  Output(UnBold);
} // TokenToHtml::ResetFormat

string TokenToHtml::DoQuoting(string& S)
{
  string Result ="";
  Result.reserve(S.length()); // it will always be at least this big -
                              // and USUALLY will be just this big
  for (int Index = 0, Count = S.length(); Index < Count; Index++) {
    unsigned char C = S[Index];
    switch (C) {
      case '<':
        Result += "&lt;";
        break;
      case '>':
        Result += "&gt;";
        break;
      case '&':
        Result += "&amp;";
        break;
      default:
        if (C < 128) Result += C;
        else Result += "&#" + IntToString(C) + ';' ;
    }
  }
  return Result;
} // TokenToHtml::DoQuoting

HGLOBAL TokensToHtml(TokenStream& Tokens)
{
  OutputList List;
  for_each(Tokens.begin(), Tokens.end(), TokenToHtml(List));
  return ConsolidateOutputList("<pre>", List, "</pre>");
} // TokensToHtml


