class TokenToRtf: public TokenToOutput
{
  public:
    TokenToRtf(OutputList& S) : TokenToOutput(S) {};
    virtual void SetFormat(Token& T);
    virtual void ResetFormat(Token& T);
    virtual string DoQuoting(string& S);
}; // TokenToRtf

#define PutOut(S)  {Output(S); OutFlag = true;}
#define PutOutV(S) {OutputV(S); OutFlag = true;}

static string Space = " ";

void TokenToRtf::SetFormat(Token& T)
{
  static string Bold      = "\\b";
  static string Italic    = "\\i";
  static string Underline = "\\ul";

  SyntaxFormat Format = SyntaxFormats[T.Type];
  bool OutFlag = false;
  if (Format.Bold)                  PutOut(Bold);
  if (Format.Italic)                PutOut(Italic);
  if (Format.Underline)             PutOut(Underline);
  if (Format.Color != clWindowText) PutOut("\\cf" + IntToString(T.Type + 1, 1));
  if (OutFlag) Output(Space);
} // TokenToRtf::SetFormat

void TokenToRtf::ResetFormat(Token& T)
{
  static string UnBold      = "\\b0";
  static string UnItalic    = "\\i0";
  static string UnUnderline = "\\ul0";
  static string UnColor     = "\\cf0";

  SyntaxFormat Format = SyntaxFormats[T.Type];
  bool OutFlag = false;
  if (Format.Bold)                  PutOut(UnBold);
  if (Format.Italic)                PutOut(UnItalic);
  if (Format.Underline)             PutOut(UnUnderline);
  if (Format.Color != clWindowText) PutOut(UnColor);
  if (OutFlag) Output(Space);
} // TokenToRtf::ResetFormat

#undef PutOut
#undef PutOutV

string TokenToRtf::DoQuoting(string& S)
{
  string Result = "";
  Result.reserve(S.length()); // it will always be at least this big - 
                              // and USUALLY will be just this big
  enum {OnLine, SeenCR, SeenLF} State = OnLine;
  for (int Index = 0, Count = S.length(); Index < Count; Index++) {
    unsigned char C = S[Index];
    #define LineBreak Result += "\\par \r\n"
/*
  if CR and not SeenLF, output linebreak; set SeenCR
  if LF and not SeenCR, output linebreak; set SeenLF
  else set Online
*/
    if (C == '\r' && State != SeenLF)      {LineBreak; State = SeenCR;}
    else if (C == '\n' && State != SeenCR) {LineBreak; State = SeenLF;}
    else {
      State = OnLine;
      switch (C) {
//        case '\r': Result += "\\par \r"; break;
        case '{':  Result += "\\{";      break;
        case '}':  Result += "\\}";      break;
        case '\\': Result += "\\\\";     break; // a special char to C and RTF ...
        default:   Result += C;
      }
    }
  }
  return Result;
} // TokenToRtf::DoQuoting

string RtfNameOfTColor(TColor Color)
{
  // Get Windows RGB value
  int RGB = ColorToRGB(Color);
  return "\\red" +   IntToString((RGB & 0x000000FF)      , 1) +
         "\\green" + IntToString((RGB & 0x0000FF00) >>  8, 1) +
         "\\blue" +  IntToString((RGB & 0x00FF0000) >> 16, 1) ;
} // RtfNameOfTColor

HGLOBAL TokensToRtf(TokenStream& Tokens)
{
  // 10 pt, Courier New, black
  string Prefix = "{\\rtf1\\ansi\\deff0\\deftab720{"
                  "\\fonttbl{\\f0\\fnil\\fcharset1 Courier New;}}\r\n"
                  "{\\colortbl\\;";
  for ( int TokenType = Low_TokenTypes;
        TokenType <= High_TokenTypes; TokenType++)
    Prefix += RtfNameOfTColor(SyntaxFormats[TokenType].Color) + ';' ;
  Prefix += "}\\deflang1033\\pard\\plain\\f0\\fs20 ";

  OutputList List;
  for_each(Tokens.begin(), Tokens.end(), TokenToRtf(List));

  return ConsolidateOutputList(Prefix, List, "\\par }");
}; // TokensToRtf


